"""A set of utilities to work with EE types."""



# Using lowercase function naming to match the JavaScript names.
# pylint: disable-msg=g-bad-name

import datetime
import numbers

import computedobject
import encodable
import geometry

# The name of the property inserted into objects created by
# ee.CustomFunction.variable() whose value is the type (class)
# of the variable.
VAR_TYPE_KEY = '__EE_VAR_TYPE'


def classToName(klass):
  """Converts a class to the API-friendly type name.

  Args:
    klass: The class.

  Returns:
    The name of the class, or "Object" if not recognized.
  """
  if issubclass(klass, computedobject.ComputedObject):
    return klass.name()
  elif klass == geometry.Geometry:
    return 'Geometry'
  elif issubclass(klass, numbers.Number):
    return 'Number'
  elif issubclass(klass, basestring):
    return 'String'
  elif issubclass(klass, (list, tuple)):
    return 'Array'
  elif issubclass(klass, datetime.datetime):
    return 'Date'
  else:
    return 'Object'


def isSubtype(firstType, secondType):
  """Checks whether a type is a subtype of another.

  Args:
    firstType: The first type name.
    secondType: The second type name.

  Returns:
    Whether secondType is a subtype of firstType.
  """
  if secondType == firstType:
    return True

  if firstType == 'EEObject':
    return secondType in ('Image', 'Feature', 'Collection', 'EECollection',
                          'ImageCollection', 'FeatureCollection')
  elif firstType in ('FeatureCollection', 'EECollection', 'Collection'):
    return secondType in ('Collection', 'EECollection',
                          'ImageCollection', 'FeatureCollection')
  elif firstType == object:
    return True
  else:
    return False


def isNumber(obj):
  """Returns true if this object is a number or number variable.

  Args:
    obj: The object to check.

  Returns:
    Whether the object is a number or number variable.
  """
  return isinstance(obj, numbers.Number) or isVarOfType(obj, numbers.Number)


def isString(obj):
  """Returns true if this object is a string or string variable.

  Args:
    obj: The object to check.

  Returns:
    Whether the object is a string or string variable.
  """
  return isinstance(obj, basestring) or isVarOfType(obj, basestring)


def isArray(obj):
  """Returns true if this object is an array or array variable.

  Args:
    obj: The object to check.

  Returns:
    Whether the object is an array or array variable.
  """
  return (isinstance(obj, (list, tuple)) or
          isVarOfType(obj, list) or
          isVarOfType(obj, tuple))


def isVarOfType(obj, klass):
  """Returns true if this object is an EE variable with the given type.

  Args:
    obj: The object to check.
    klass: The class to check against.

  Returns:
    Whether the object is a variable of the given type.
  """
  if isinstance(obj, encodable.Encodable) and hasattr(obj, VAR_TYPE_KEY):
    actual_klass = getattr(obj, VAR_TYPE_KEY)
    return actual_klass is not None and issubclass(actual_klass, klass)
  else:
    return False



"""Common representation for ImageCollection and FeatureCollection.

This class is never intended to be instantiated by the user.
"""



# Using lowercase function naming to match the JavaScript names.
# pylint: disable=g-bad-name

import apifunction
import computedobject
import customfunction
import ee_exception
import ee_types
import filter   # pylint: disable=redefined-builtin
import function


class Collection(computedobject.ComputedObject):
  """Base class for ImageCollection and FeatureCollection."""

  _initialized = False

  # The serial number of the next mapping variable.
  _serialMappingId = 0

  def __init__(self, func, args):
    """Constructs a collection by initializing its ComputedObject."""
    super(Collection, self).__init__(func, args)

  @classmethod
  def initialize(cls):
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, 'Collection', 'Collection')
      apifunction.ApiFunction.importApi(
          cls, 'AggregateFeatureCollection', 'Collection', 'aggregate_')
      cls._initialized = True

  @classmethod
  def reset(cls):
    """Removes imported API functions from this class.

    Also resets the serial ID used for mapping Python functions to 0.
    """
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False
    cls._serialMappingId = 0

  def filter(self, new_filter):
    """Apply a filter to this collection.

    Args:
      new_filter: Filter to add to this collection.

    Returns:
      The filtered collection object.
    """
    if not new_filter:
      raise ee_exception.EEException('Empty filters.')
    return self._cast(apifunction.ApiFunction.call_(
        'Collection.filter', self, new_filter))

  def filterMetadata(self, name, operator, value):
    """Shortcut to add a metadata filter to a collection.

    This is equivalent to self.filter(Filter().metadata(...)).

    Args:
      name: Name of a property to filter.
      operator: Name of a comparison operator as defined
          by FilterCollection.  Possible values are: "equals", "less_than",
          "greater_than", "not_equals", "not_less_than", "not_greater_than",
          "starts_with", "ends_with", "not_starts_with", "not_ends_with",
          "contains", "not_contains".
      value: The value to compare against.

    Returns:
      The filtered collection.
    """
    return self.filter(filter.Filter.metadata_(name, operator, value))

  def filterBounds(self, geometry):
    """Shortcut to add a geometry filter to a collection.

    Items in the collection with a footprint that fails to intersect
    the given geometry will be excluded when the collection is evaluated.
    This is equivalent to self.filter(Filter().geometry(...)).

    Args:
      geometry: The boundary to filter to either as a GeoJSON geometry,
          or a FeatureCollection, from which a geometry will be extracted.

    Returns:
      The filter object.
    """
    return self.filter(filter.Filter.geometry(geometry))

  def filterDate(self, start, opt_end=None):
    """Shortcut to filter a collection with a date range.

    Items in the collection with a time_start property that doesn't
    fall between the start and end dates will be excluded.
    This is equivalent to self.filter(Filter().date(...)).

    Args:
      start: The start date as a Date object, a string representation of
          a date, or milliseconds since epoch.
      opt_end: The end date as a Date object, a string representation of
          a date, or milliseconds since epoch.

    Returns:
      The filter object.
    """
    return self.filter(filter.Filter.date(start, opt_end))

  def getInfo(self):
    """Returns all the known information about this collection.

    This function makes an REST call to to retrieve all the known information
    about this collection.

    Returns:
      The return contents vary but will include at least:
       features: an array containing metadata about the items in the
           collection that passed all filters.
       properties: a dictionary containing the collection's metadata
           properties.
    """
    return super(Collection, self).getInfo()

  def limit(self, maximum, opt_property=None, opt_ascending=None):
    """Limit a collection to the specified number of elements.

    This limits a collection to the specified number of elements, optionally
    sorting them by a specified property first.

    Args:
       maximum: The number to limit the collection to.
       opt_property: The property to sort by, if sorting.
       opt_ascending: Whether to sort in ascending or descending order.
           The default is true (ascending).

    Returns:
       The collection.
    """
    args = {'collection': self, 'limit': maximum}
    if opt_property is not None:
      args['key'] = opt_property
    if opt_ascending is not None:
      args['ascending'] = opt_ascending
    return self._cast(
        apifunction.ApiFunction.apply_('Collection.limit', args))

  def sort(self, prop, opt_ascending=None):
    """Sort a collection by the specified property.

    Args:
       prop: The property to sort by.
       opt_ascending: Whether to sort in ascending or descending
           order.  The default is true (ascending).

    Returns:
       The collection.
    """
    args = {'collection': self, 'key': prop}
    if opt_ascending:
      args['ascending'] = opt_ascending
    return self._cast(
        apifunction.ApiFunction.apply_('Collection.limit', args))

  def geometry(self):
    """Run an algorithm to extract the geometry from this collection.

    Returns:
      A naked Algorithm invocation (a JSON dictionary), not an ee object,
      however this object can be used any place a Geometry object can be used.
    """
    return apifunction.ApiFunction.call_('ExtractGeometry', self)

  @classmethod
  def _cast(cls, obj):
    """Cast a ComputedObject to a new instance of the same class as this.

    Args:
      obj: The object to cast.

    Returns:
      The cast object, and instance of the class on which this method is called.
    """
    if isinstance(obj, cls):
      return obj
    else:
      # Assumes all subclass constructors can be called with a
      # ComputedObject as their first parameter.
      return cls(obj)

  @staticmethod
  def name():
    return 'Collection'

  def mapInternal(self,
                  cls,
                  algorithm,
                  opt_dynamicArgs=None,
                  opt_constantArgs=None,
                  opt_destination=None):
    """Maps an algorithm over a collection.

    Args:
      cls: The collection elements' type (class).
      algorithm: The operation to map over the images or features of the
          collection. Either an algorithm name as a string, or a Python
          function that receives an image or features and returns one. If a
          function is passed, it is called only once and the result is captured
          as a description, so it cannot perform imperative operations or rely
          on external state.
      opt_dynamicArgs: A map specifying which properties of the input objects
          to pass to each argument of the algorithm. This maps from argument
          names to selector strings. Selector strings are property names,
          optionally concatenated into chains separated by a period to access
          properties-of-properties. To pass the whole object, use the special
          selector string '.all', and to pass the geometry, use '.geo'. If
          this argument is not specified, the names of the arguments will be
          matched exactly to the properties of the input object. If algorithm
          is a Python function, this must be null or undefined as the
          image will always be the only dynamic argument.
      opt_constantArgs: A map from argument names to constant values to be
          passed to the algorithm on every invocation.
      opt_destination: The property where the result of the algorithm will be
          put. If this is null or undefined, the result of the algorithm will
          replace the input, as is the usual behavior of a mapping opeartion.

    Returns:
      The mapped collection.

    Raises:
      RuntimeError: if algorithm is a Python function and opt_dynamicArgs is
          specified.
    """
    if callable(algorithm):
      if opt_dynamicArgs:
        # TODO(user): Remove this once we have a getProperty() algorithm.
        raise ee_exception.EEException(
            'Can\'t use dynamicArgs with a mapped Python function.')
      varName = '_MAPPING_VAR_%d' % Collection._serialMappingId
      Collection._serialMappingId += 1
      algorithm = customfunction.CustomFunction({varName: cls}, cls, algorithm)
    elif isinstance(algorithm, basestring):
      algorithm = apifunction.ApiFunction.lookup(algorithm)
    elif not isinstance(algorithm, function.Function):
      raise ee_exception.EEException(
          'Can\'t map non-callable object: %s' % algorithm)

    args = {
        'collection': self,
        'baseAlgorithm': algorithm
    }
    if opt_dynamicArgs:
      args['dynamicArgs'] = opt_dynamicArgs
    else:
      # Use the function's first argument.
      varName = algorithm.getSignature()['args'][0]['name']
      args['dynamicArgs'] = {varName: '.all'}

    if opt_constantArgs: args['constantArgs'] = opt_constantArgs
    if opt_destination: args['destination'] = opt_destination
    return self._cast(apifunction.ApiFunction.apply_('Collection.map', args))

  @classmethod
  def createAutoMapFunctions(cls, elementClass):
    """Creates map_*() methods on collectionClass.

    Creates a map_* method on collectionClass for each generated instance
    method on elementClass that maps that method over the collection.

    TODO(user): Deprecate these.

    Args:
      elementClass: The collection elements' type.
    """
    for name in dir(elementClass):
      method = getattr(elementClass, name)
      if callable(method) and hasattr(method, 'signature'):
        def MakeMapFunction(name, method):   # Capture scope.
          def MapFunction(self, *args, **kwargs):
            method_caller = lambda elem: method(elem, *args, **kwargs)
            destination = None
            if not ee_types.isSubtype('EEObject', method.signature['returns']):
              destination = name
            return self.mapInternal(
                elementClass, method_caller, None, None, destination)
          return MapFunction
        setattr(cls, 'map_' + name, MakeMapFunction(name, method))



"""A representation of an Earth Engine computed object."""



# Using lowercase function naming to match the JavaScript names.
# pylint: disable=g-bad-name

import data
import encodable
import serializer


class ComputedObject(encodable.Encodable):
  """A representation of an Earth Engine computed object.

  This is a base class for most API objects.

  The class itself is not abstract as it is used to wrap the return values of
  algorithms that produce unrecognized types with the minimal functionality
  necessary to interact well with the rest of the API.
  """

  class __metaclass__(type):
    """A meta-class that makes type coercion idempotent.

    If an instance of a ComputedObject subclass is instantiated by passing
    another instance of that class as the sole argument, this short-circuits
    and returns that argument.

    This is necessary for CustomFunction variables to pretend to be instances
    of any type.
    """

    def __call__(cls, *args, **kwargs):
      """Creates a computed object, catching self-casts."""
      if len(args) == 1 and not kwargs and isinstance(args[0], cls):
        # Self-casting returns the argument unchanged.
        return args[0]
      else:
        return type.__call__(cls, *args, **kwargs)

  def __init__(self, func, args):
    """Creates a computed object.

    Args:
      func: The ee.Function called to compute this object, either as an
          Algorithm name or an ee.Function object.
      args: A dictionary of arguments to pass to the specified function.
          Note that the caller is responsible for promoting the arguments
          to the correct types.
    """
    self.func = func
    self.args = args

  def __eq__(self, other):
    return (type(self) == type(other) and
            self.func == other.func and
            self.args == other.args)

  def __ne__(self, other):
    return not self.__eq__(other)

  def getInfo(self):
    """Fetch and return information about this object.

    Returns:
      The object can evaluate to anything.
    """
    return data.getValue({'json': self.serialize()})

  def encode(self, encoder):
    """Encodes the object in a format compatible with Serializer."""

    # Encode the function that we're calling.
    func = encoder(self.func)
    # Built-in functions are encoded as strings under a different key.
    key = 'functionName' if isinstance(func, basestring) else 'function'

    # Encode all arguments recursively.
    encoded_args = {}
    for name, value in self.args.iteritems():
      if value is not None:
        encoded_args[name] = encoder(value)

    return {
        'type': 'Invocation',
        'arguments': encoded_args,
        key: func
    }

  def serialize(self, opt_pretty=False):
    """Serialize this object into a JSON string.

    Args:
      opt_pretty: A flag indicating whether to pretty-print the JSON.

    Returns:
      The serialized representation of this object.
    """
    return serializer.toJSON(self, opt_pretty)

  def __str__(self):
    """Writes out the object in a human-readable form."""
    return 'ee.%s(%s)' % (self.name(), serializer.toReadableJSON(self))

  @classmethod
  def name(cls):
    """Returns the name of the object, used in __str__()."""
    return 'ComputedObject'



"""A serializer that encodes EE object trees as JSON DAGs."""



# Using lowercase function naming to match the JavaScript names.
# pylint: disable=g-bad-name

import datetime
import json
import math
import numbers

import ee_exception
import encodable


# The datetime for the beginning of the Unix epoch.
_EPOCH_DATETIME = datetime.datetime.utcfromtimestamp(0)


class Serializer(object):
  """A serializer for EE object trees."""

  def __init__(self, is_compound=True):
    """Constructs a serializer.

    Args:
      is_compound: Whether the encoding should factor out shared subtrees.
    """

    # Whether the encoding should factor out shared subtrees.
    self._is_compound = bool(is_compound)

    # A list of shared subtrees as [name, value] pairs.
    self._scope = []

    # A lookup table from object ID as retrieved by id() to subtree
    # names as stored in self._scope.
    self._encoded = {}

  def _encode(self, obj):
    """Encodes a top level object in the EE API v2 (DAG) format.

    Args:
      obj: The object to encode.

    Returns:
      An encoded object ready for JSON serialization.
    """
    value = self._encodeValue(obj)
    if self._is_compound:
      if (isinstance(value, dict) and
          value['type'] == 'ValueRef' and
          len(self._scope) == 1):
        # Just one value. No need for complex structure.
        value = self._scope[0][1]
      else:
        # Wrap the scopes and final value with a CompoundValue.
        value = {
            'type': 'CompoundValue',
            'scope': self._scope,
            'value': value
        }
      # Clear state in case of future encoding.
      self._scope = []
      self._encoded = {}
    return value

  def _encodeValue(self, obj):
    """Encodes a subtree as a Value in the EE API v2 (DAG) format.

    If _is_compound is True, this will fill the _scope and _encoded properties.

    Args:
      obj: The object to encode.

    Returns:
      An encoded object.
    """
    obj_id = id(obj)
    encoded = self._encoded.get(obj_id, None)

    if encoded is not None:
      # Already encoded objects are encoded as ValueRefs and returned directly.
      return {
          'type': 'ValueRef',
          'value': encoded
      }
    elif obj is None or isinstance(obj, (bool, numbers.Number, basestring)):
      # Primitives are encoded as is and not saved in the scope.
      return obj
    elif isinstance(obj, datetime.datetime):
      # Dates are encoded as typed UTC microseconds since the Unix epoch.
      # They are returned directly and not saved in the scope either.
      microseconds = (obj - _EPOCH_DATETIME).total_seconds() * 1e6
      return {
          'type': 'Date',
          'value': math.floor(microseconds)
      }
    elif isinstance(obj, encodable.Encodable):
      # Some objects know how to encode themselves.
      result = obj.encode(self._encodeValue)
      if not isinstance(result, dict) or result['type'] == 'ArgumentRef':
        # Optimization: simple enough that adding it to the scope is probably
        # not worth it.
        return result
    elif isinstance(obj, (list, tuple)):
      # Lists are encoded recursively.
      result = [self._encodeValue(i) for i in obj]
    elif isinstance(obj, dict):
      # Dictionary are encoded recursively and wrapped in a type specifier.
      result = {
          'type': 'Dictionary',
          'value': {key: self._encodeValue(value) for key, value in obj.items()}
      }
    else:
      raise ee_exception.EEException('Can\'t encode object: %s' % obj)

    if self._is_compound:
      # Save the new object and return a ValueRef.
      name = str(len(self._scope))
      self._scope.append((name, result))
      self._encoded[obj_id] = name
      return {
          'type': 'ValueRef',
          'value': name
      }
    else:
      return result


def toJSON(obj, opt_pretty=False):
  """Serialize an object to a JSON string appropriate for API calls.

  Args:
    obj: The object to Serialize.
    opt_pretty: True to pretty-print the object.

  Returns:
    A JSON string representing the input.
  """
  serializer = Serializer(not opt_pretty)
  encoded = serializer._encode(obj)  # pylint: disable=protected-access
  return json.dumps(encoded, indent=2 if opt_pretty else None)


def toReadableJSON(obj):
  """Convert an object to readable JSON."""
  return toJSON(obj, True)



"""An object representing a custom EE Function."""



# Using lowercase function naming to match the JavaScript names.
# pylint: disable-msg=g-bad-name

import ee_exception
import ee_types
import encodable
import function


class CustomFunction(function.Function):
  """An object representing a custom EE Function."""

  def __init__(self, args, return_type, body):
    """Creates a function defined by a given expression with unbound variables.

    Args:
      args: The arguments to the function. If body is a Python function, this
          must be a map from argument names to types (classes). Otherwise this
          is an array of names.
      return_type: The expected return type of the function as a class,
          or None if not known.
      body: The expression to evaluate. Can be either:
          1. A Python function, in which case the args argument must describe
             the types of this function's arguments.
          2. An encodable value, made out of primitives (boolean, number,
             string, datetime), data structures (lists, dictionaries),
             ComputedObjects or variables (made by CustomFunction.variable()).
             Variables refer to the elements of args by names. Structures and
             ComputedObjects can recursively contain any of the above.
    """
    if callable(body):
      if not isinstance(args, dict):
        raise ee_exception.EEException(
            'The "args" of a custom function created from a native '
            'Python function must be a dictionary from name to type.')
      arg_names = args.keys()
      arg_types = args.values()
      variables = [CustomFunction.variable(var_type, name)
                   for name, var_type in args.iteritems()]
      body = body(*variables)
    else:
      if not isinstance(args, (list, tuple)):
        raise ee_exception.EEException(
            'The "args" of a custom function created from an '
            'expression must be a list or tuple of names.')
      arg_names = args
      arg_types = [object] * len(args)

    # The names of the function arguments.
    self._arg_names = arg_names

    # The types of the function arguments.
    self._arg_types = arg_types

    # The return type of the function.
    self._return_type = return_type or object

    # The expression to evaluate.
    self._body = body

  def encode(self, encoder):
    return {
        'type': 'Function',
        'argumentNames': self._arg_names,
        'body': encoder(self._body)
    }

  def getSignature(self):
    """Returns a description of the interface provided by this function."""
    return {
        'name': '',
        'args': [
            {'name': name, 'type': ee_types.classToName(cls), 'optional': False}
            for name, cls in zip(self._arg_names, self._arg_types)
        ],
        'returns': ee_types.classToName(self._return_type)
    }

  @staticmethod
  def variable(var_type, name):
    """Returns a placeholder variable with a given name and EE type.

    Args:
      var_type: A class to mimic.
      name: The name of the variable as it will appear in the
          arguments of the custom functions that use this variable.

    Returns:
      A variable with the given name implementing the given type.
    """
    var_type = var_type or object
    if issubclass(var_type, encodable.Encodable):
      base = var_type
    else:
      base = encodable.Encodable

    class Variable(base):
      def __init__(self):
        pass  # Don't call the base class's constructor.

      def encode(self, unused_encoder):
        return {
            'type': 'ArgumentRef',
            'value': name
        }

    instance = Variable()
    setattr(instance, ee_types.VAR_TYPE_KEY, var_type)
    return instance
